(* $Id: eval.ml,v 1.8 2021/04/27 02:57:53 garrigue Exp $ *)

open StdLabels
open Terms

module Env = Map.Make(struct type t = string let compare = compare end)
let get ~env x =
  try Env.find x env
  with Not_found -> Nil

let vtrue = Cons(Nil,Nil)

let rec eval_expr ~env = function
    Var x -> get ~env x
  | Cons(Atom"cons", Cons(e1,Cons(e2,Nil))) ->
      Cons(eval_expr ~env e1, eval_expr ~env e2)
  | Cons(Atom("hd"|"tl" as s), Cons(e,Nil)) ->
      begin match eval_expr ~env e with
        Cons(h,t) -> if s = "hd" then h else t
      | _ -> Nil
      end
  | Cons(Atom"=?", Cons(e1,Cons(e2,Nil))) ->
      if eval_expr ~env e1 = eval_expr ~env e2 then vtrue else Nil
  | v -> v

let trace_fun = ref (fun ~env cmd -> ())

let rec bind_args env xs v =
  match xs, v with
  | [x], v -> Env.add x v env
  | x :: xs, Cons (v, w) -> bind_args (Env.add x v env) xs w
  | _ -> env

let rec eval_cmd ~procs ~env cmd =
  !trace_fun ~env cmd;
  match cmd  with
    Assign (x, e) ->
      Env.add x (eval_expr ~env e) env
  | Seq (c1, c2) ->
      eval_cmd c2 ~env:(eval_cmd c1 ~env ~procs) ~procs
  | Skip ->
      env
  | While (e, c1) as c ->
      if eval_expr ~env e = Nil then env else
      eval_cmd c ~env:(eval_cmd c1 ~env ~procs) ~procs
  | Call (x, f, e) ->
      Env.add x (eval ~procs f (eval_expr ~env e)) env

and eval ~procs f v =
  let (xs,c,y) =
    try Env.find f procs
    with Not_found -> failwith ("Unknown procedure " ^ f)
  in
  let env = bind_args Env.empty xs v in
  let env = eval_cmd ~env ~procs c in
  get ~env y

(*
let reverse =
  ("X",
   Seq(Assign("Y",Val Nil),
       While(Var"X", Seq(Assign("Y", Econs(Ehd(Var"X"), Var"Y")),
                      Assign("X", Etl(Var"X"))))),
   "Y")

let eval_string p v =
  eval (program_of_string p) (value_of_string v)
*)
